import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:typed_data';

import 'package:http/http.dart';
import 'package:http/io_client.dart';
import 'package:logger/logger.dart';
import 'package:wakatime_api/wakatime_api.dart';

/// Si occupa di effettuare le chiamate REST
/// e di inviare i dati di FirebasePerformance
class Provider {
  // ignore: public_member_api_docs
  factory Provider() => _instance;

  Provider._internal();

  static final Provider _instance = Provider._internal();

  late Client _client;

  /// Inizializza il [Provider] impostando anche il certificato
  /// per il [SecurityContext]
  Future<void> init() async {
    _client = IOClient(
      HttpClient()
        ..badCertificateCallback = (certificate, host, port) {
          if (host == 'maps.googleapis.com') return true;
          return false;
        },
    );
    logger.i('[Provider] Successfully initialized');
  }

  Future<void> _printLog(String call, Response response, String method) async {
    String callUrl = call;

    /// Stampa tutto l'url della chiamata tranne i query params
    if (callUrl.contains('?')) {
      callUrl = callUrl.substring(0, callUrl.indexOf('?'));
    }
    late Level level;

    switch (response.statusCode) {
      case 200:
        level = Level.info;
        break;
      case 500:
        level = Level.error;
        break;
      default:
        level = Level.warning;
        break;
    }

    var responseData =
        // ignore: lines_longer_than_80_chars
        '$method: ${response.statusCode} - ${response.reasonPhrase} - response of: ${response.statusCode == 404 ? 0 : response.body.length} bytes';

    if (response.statusCode != 200) {
      responseData = '$responseData -- ${response.body}';
    }

    if (level == Level.warning) {
      logger.w('[Provider-meh] $callUrl: $responseData');
    } else if (level == Level.error) {
      logger.w('[Provider-nope] $callUrl: $responseData');
    } else {
      logger.i('[Provider-aight] $callUrl: $responseData');
    }
  }

  /// Effettua una post
  ///
  /// Solo uno fra `data` e `dataAsList` può essere specificato
  Future<Response> post(
    String url, {
    Map? data,
    String tokenName = 'Authorization',
    List? dataAsList,
    Map<String, String>? headers,
  }) async {
    assert(
      (dataAsList != null && data == null) ||
          (dataAsList == null && data != null) ||
          (dataAsList == null && data == null),
      "Solo uno fra 'dataAsList' e 'data' può essere specificato",
    );

    (headers ??= <String, String>{})
        .putIfAbsent('Content-Type', () => 'application/json');

    try {
      final uri = Uri.parse(url);

      late String body;

      if (dataAsList != null) {
        body = jsonEncode(dataAsList);
      } else if (data != null) {
        body = jsonEncode(data);
      } else {
        body = jsonEncode({});
      }

      final response = await _client.post(
        uri,
        body: body,
        headers: headers,
      );

      _printLog(url, response, 'POST');
      return response;
    } catch (e) {
      logger.e(
        // ignore: lines_longer_than_80_chars
        '[Provider-POST] Error on url $url: ${e.toString().replaceAll("\n", ' ')}',
      );
      return Response(jsonEncode({'code': 'GENERIC_ERROR'}), 500);
    }
  }

  /// Effettua una chiamata `GET`
  Future<Response> get(
    String url, {
    Map<String, String>? headers,
    String Function(Uint8List)? responseDecoder,
    Map<String, String>? queryParams,
  }) async {
    String urlToCall = url;

    try {
      /// Aggiunge i queryParams
      if (queryParams != null && queryParams.isNotEmpty) {
        // ignore: parameter_assignments
        urlToCall =
            // ignore: lines_longer_than_80_chars
            "$url?${queryParams.entries.map((e) => "${e.key}=${e.value}").reduce((value, element) => "$value&$element")}";
      }

      final response =
          await _client.get(Uri.parse(urlToCall), headers: headers);

      if (responseDecoder != null) {
        return Response(
          responseDecoder.call(response.bodyBytes),
          response.statusCode,
        );
      }

      _printLog(urlToCall, response, 'GET');
      return response;
    } catch (e) {
      logger.w('[Provider-GET] Error on url $urlToCall: $e');
      return Response(jsonEncode({'code': 'GENERIC_ERROR'}), 500);
    }
  }

  /// Manda una `PATCH`
  Future<Response> patch(
    String url, {
    Map? data,
    Map<String, String>? headers,
  }) async {
    (headers ??= <String, String>{})
        .putIfAbsent('Content-Type', () => 'application/json');

    try {
      final uri = Uri.parse(url);

      final response = await _client.patch(
        uri,
        body: data == null ? {} : jsonEncode(data),
        headers: headers,
      );

      _printLog(url, response, 'PATCH');
      return response;
    } catch (e) {
      logger.w('[Provider-PATCH] Error on url $url: $e');
      return Response(jsonEncode({'code': 'GENERIC_ERROR'}), 500);
    }
  }

  /// Manda una `DELETE`
  Future<Response> delete(
    String url, {
    Map? data,
    Map<String, String>? headers,
  }) async {
    (headers ??= <String, String>{})
        .putIfAbsent('Content-Type', () => 'application/json');

    try {
      final uri = Uri.parse(url);

      final response = await _client.delete(
        uri,
        body: data == null ? {} : jsonEncode(data),
        headers: headers,
      );

      _printLog(url, response, 'DELETE');
      return response;
    } catch (e) {
      logger.w('[Provider-DELETE] Error on url $url: $e');
      return Response(jsonEncode({'code': 'GENERIC_ERROR'}), 500);
    }
  }
}
