import 'dart:convert';

import 'package:flutter/foundation.dart';
import 'package:intl/intl.dart';
import 'package:wakatime_api/apis/provider.dart';
import 'package:wakatime_api/models/aggregated_data.dart';
import 'package:wakatime_api/models/duration_mixin.dart';
import 'package:wakatime_api/models/project.dart';
import 'package:wakatime_api/models/slice_by.dart';
import 'package:wakatime_api/wakatime_api.dart';

class WakatimeApi {
  // ignore: public_member_api_docs
  factory WakatimeApi([String? apiKey]) => _instance ??= WakatimeApi._(apiKey!);

  WakatimeApi._(String _apiKey) {
    _defaultHeaders.putIfAbsent('Authorization', () {
      final token = base64Encode(utf8.encode(_apiKey));
      return 'Basic $token';
    });
  }
  static WakatimeApi? _instance;

  final _provider = Provider()..init();

  /// Base url of wakatime
  final _baseUrl = 'https://wakatime.com/api/v1';

  final _defaultHeaders = <String, String>{};

  /// Gets the heartbeats for a given [date]
  Future<List<Heartbeat>> getHeartbeats(DateTime date) async {
    final path = '$_baseUrl/users/current/heartbeats';
    final parsedDate = DateFormat('yyyy-MM-dd').format(date);
    logger.d('Getting heartbeats for date $parsedDate');
    final response = await _provider.get(
      path,
      queryParams: {'date': parsedDate},
      headers: _defaultHeaders,
    );

    if (response.statusCode == 200) {
      final data = jsonDecode(response.body) as Map<String, dynamic>;
      return (data['data'] as List<dynamic>)
          .map((e) => Heartbeat.fromJson(e))
          .toList();
    } else {
      return [];
    }
  }

  /// Gets the heartbeats for the range between [initialDate] and
  /// [finalDate]
  Future<List<Heartbeat>> getHeartBeatsForRange(
    DateTime initialDate,
    DateTime finalDate,
  ) async {
    assert(initialDate.isBefore(finalDate));

    final dates = List.generate(
      initialDate.difference(finalDate).inDays.abs() + 1,
      (index) => initialDate.add(Duration(days: index)),
    );

    final heartbeats = <Heartbeat>[];

    await for (final date in Stream.fromIterable(dates)) {
      heartbeats.addAll(await getHeartbeats(date));
    }

    heartbeats.sort(
      (h1, h2) =>
          DateTime.parse(h1.createdAt).compareTo(DateTime.parse(h2.createdAt)),
    );

    return heartbeats;
  }

  /// Gets the project connected to the current user
  Future<List<Project>> getProjects() async {
    final path = '$_baseUrl/users/current/projects';
    final response = await _provider.get(
      path,
      headers: _defaultHeaders,
    );
    if (response.statusCode == 200) {
      final data =
          (jsonDecode(response.body) as Map<String, dynamic>)['data'] as List;
      return data.map((e) => Project.fromJson(e)).toList();
    }
    return [];
  }

  /// Gets the time entries for the current user
  ///
  /// If [date] is not defined, it will be `DateTime.now()`
  Future<List<HeartbeatDuration>> getDurations({
    DateTime? date,
    String? project,
    SlicedBy? sliceBy,
  }) async {
    date ??= DateTime.now();
    final path = '$_baseUrl/users/current/durations';
    final response = await _provider.get(
      path,
      headers: _defaultHeaders,
      queryParams: {
        'date': DateFormat('yyyy-MM-dd').format(date),
        if (project != null) 'project': project,
        if (sliceBy != null) 'slice_by': sliceBy.value,
      },
    );
    if (response.statusCode == 200) {
      final data = (jsonDecode(response.body) as Map<String, dynamic>);
      return (data['data'] as List)
          .map((i) => HeartbeatDuration.parse(i))
          .toList();
    }
    return [];
  }

  /// Gets the stats from day 0 since today for the current user
  Future<AggregatedData?> getAllTimeSinceToday({
    String? project,
  }) async {
    final path = '$_baseUrl/users/current/all_time_since_today';
    final response = await _provider.get(
      path,
      headers: _defaultHeaders,
      queryParams: {
        if (project != null) 'project': project,
      },
    );
    if (response.statusCode == 200) {
      final data = (jsonDecode(response.body) as Map<String, dynamic>);
      return AggregatedData.fromJson(data['data'] as Map<String, dynamic>);
    }
    return null;
  }
}
